-----------------------------------------------------------------------------
--
-- Expression->Factor are currenetly shortcut (not implemented)
--
-- Tag				:= { Statement } @
-- Statement		:= literal | "[" RValue "]"
-- RValue			:= Relation
-- Relation			:= Expression [ ("==" | "~=" | ">" | "<" | "<=" | ">=") Expression ]
-- Expression		:= Term [ ("+" | "-") Term ]
-- Factor			:= Primary [ ("*" | "/") Primary ]
-- Chain			:= Primary { ":" Primary }
-- Primary			:= ConstantInteger | ConstantString | Identifier ["(" ParameterList ")"]
-- ParameterList	:= RValue {, RValue }
--
-----------------------------------------------------------------------------

AloftTag = { }
AloftTagMethods = { }
AloftTagDirectAccessData = { isBoss = true, level = true, name = true, health = true, maxHealth = true, healthBarValue = true, healthBarMaxValue = true }
AloftTagData = 
{ 
	health 			= { events = "Aloft:OnHealthBarValueChanged"},
	healthBarValue 	= { events = "Aloft:OnHealthBarValueChanged"},
	maxHealth 		= { events = "Aloft:OnHealthBarValueChanged"},
}

-----------------------------------------------------------------------------

AloftTagMethods.angle = function(aloftData, value) return value and value ~= "" and string.format("<%s>", value) or nil end
AloftTagMethods.bracket = function(aloftData, value) return value and value ~= "" and string.format("(%s)", value) or nil end
AloftTagMethods.brace = function(aloftData, value) 	return value and value ~= "" and string.format("{%s}", value) or nil end

AloftTagMethods.healthBarDeficit = function(aloftData, value)
	return value and (aloftData.healthBarMaxValue - aloftData.healthBarValue) or nil
end
AloftTagData.healthBarDeficit = { events = "Aloft:OnHealthBarValueChanged" }

function AloftTagMethods.healthDeficit(aloftData, value)
	if value and aloftData.health and aloftData.maxHealth then
		return aloftData.maxHealth - aloftData.health
	else
		return nil
	end
end
AloftTagData.healthDeficit = { events = "Aloft:OnHealthBarValueChanged" }

AloftTagMethods.healthPercent = function(aloftData, value)
	return value and floor(100*aloftData.healthBarValue / aloftData.healthBarMaxValue) or nil
end
AloftTagData.healthPercent = { events = "Aloft:OnHealthBarValueChanged" }

function AloftTagMethods.hideKnownHealth(aloftData, value)
	return not aloftData.health and value or nil
end

AloftTagMethods.isGroup = function(aloftData, value)
	return aloftData.healthBarMaxValue ~= 100 and value or nil
end

AloftTagMethods.isNotGroup = function(aloftData, value)
	return aloftData.healthBarMaxValue == 100 and value or nil
end

AloftTagMethods.hexColor = function(aloftData, value, colorValue)
	return value and string.format("|cff%s%s|r", colorValue, value or "") or nil
end

AloftTagMethods.hideIf = function(aloftData, value, condition)
	return not condition and value or nil
end

AloftTagMethods.hideZero = function(aloftData, value)
	return value ~= 0 and value or nil
end

AloftTagMethods.isFriendly = function(aloftData, value)
	return (aloftData.type == "friendlyPlayer" or aloftData.type == "friendlyNPC") and value
end

AloftTagMethods.isFriendlyNPC = function(aloftData, value)
	return aloftData.type == "friendlyPlayer" and value
end

AloftTagMethods.isFriendlyPlayer = function(aloftData, value)
	return aloftData.type == "friendlyPlayer" and value
end

AloftTagMethods.isHostile = function(aloftData, value)
	return aloftData.type == "hostile" and value
end

AloftTagMethods.isNeutral = function(aloftData, value)
	return aloftData.type == "neutral" and value
end

AloftTagMethods.isMaxHealth = function(aloftData, value)
	return aloftData.healthBarValue == aloftData.healthBarMaxValue and value or nil
end
AloftTagData.isMaxHealth = { events = "Aloft:OnHealthBarValueChanged" }

AloftTagMethods.lower = function(aloftData, value)
	if type(value) == "string" then
		return value:lower()
	end
	return value
end

AloftTagMethods.negate = function(aloftData, value)
	if type(value) == "number" then
		return -value
	end
	return value
end

AloftTagMethods["not"] = function(aloftData, value, condition)
	return value and not condition
end

AloftTagMethods.percent = function(aloftData, value)
	if type(value) == "number" then
		return value.."%"
	end
	return value
end

AloftTagMethods.prefix = function(aloftData, value, prefix)
	return value and prefix..value or nil
end

AloftTagMethods.short = function(aloftData, value)
	if type(value) == "number" then
		if value < 10000 then
			-- Force it to an integer
			return string.format("%d", value)
		elseif value < 100000 then
			return string.format("%.1fk", value / 1000)
		elseif value < 1000000 then
			return string.format("%dk", value / 1000)
		else
			return string.format("%.2fm", value / 1000000)
		end
	end
	return value
end

AloftTagMethods.suffix = function(aloftData, value, suffix)
	return value and value..suffix or nil
end

AloftTagMethods.upper = function(aloftData, value)
	if type(value) == "string" then
		return value:upper()
	end
	return value
end

AloftTagMethods.white = function(aloftData, value) return value and ("|r%s"):format(value) or nil end
AloftTagMethods.red = function(aloftData, value) return value and ("|cffff0000%s|r"):format(value) or nil end
AloftTagMethods.green = function(aloftData, value) return value and ("|cff00ff00%s|r"):format(value) or nil end
AloftTagMethods.blue = function(aloftData, value) return value and ("|cff0000ff%s|r"):format(value) or nil end
AloftTagMethods.cyan = function(aloftData, value) return value and ("|cff00ffff%s|r"):format(value) or nil end
AloftTagMethods.fuchsia = function(aloftData, value) return value and ("|cffff00ff%s|r"):format(value) or nil end
AloftTagMethods.yellow = function(aloftData, value) return value and ("|cffffff00%s|r"):format(value) or nil end
AloftTagMethods.gray = function(aloftData, value) return value and ("|cffafafaf%s|r"):format(value) or nil end
AloftTagMethods.difficultyColor = function(aloftData, value) return value and ("|cff%02x%02x%02x%s|r"):format(aloftData.levelTextR*255, aloftData.levelTextG*255, aloftData.levelTextB*255, value) or nil end

-----------------------------------------------------------------------------

local function DataTableGetByte(this) while this.tag:byte(this.offset) == 32 do this.offset = this.offset + 1 end return this.tag:byte(this.offset) or 0 end
local function DataTableAdvanceByte(this, offset) this.offset = this.offset + 1 end
local function EmptyStringMethod() return "" end
local function TagError(x) Aloft:Print(x) end

function AloftTag:ParseParameterList(dataTable)
	if dataTable:GetByte() == 40 then -- byte == '('
		dataTable:AdvanceByte()
		local byte = dataTable:GetByte()
		if byte == 41 then 
			dataTable:AdvanceByte()
			return ""
		end

		local parameterListString = string.format(", %s", self:ParseRValue(dataTable))
		
		while dataTable.offset <= #dataTable.tag and dataTable:GetByte() == 44 do -- byte == ','
			dataTable:AdvanceByte()
			parameterListString = string.format("%s, %s", parameterListString, self:ParseRValue(dataTable))
		end

		if dataTable:GetByte() == 41 then -- byte == ')'
			dataTable:AdvanceByte()
		else
			dataTable.error = true
			TagError("Missing end of parameter list \")\"")
		end
		
		return parameterListString
	else
		return ""
	end
end

function AloftTag:ParsePrimary(dataTable, functionString)
	-- Either an integer, a string or an LValue (value or function call)
	local byte = dataTable:GetByte()
	if byte >= 48 and byte <= 59 then -- byte >= '0' && byte <= '9'
		local startOffset = dataTable.offset
		local endOffset = startOffset
		repeat
			endOffset = endOffset + 1
			byte = dataTable.tag:byte(endOffset) or 0
		until byte < 48 or byte > 59
		dataTable.offset = endOffset
		if functionString then
			return string.format("(%s and %s or nil)", functionString, dataTable.tag:sub(startOffset, endOffset-1))
		else
			return dataTable.tag:sub(startOffset, endOffset-1)
		end
	elseif byte == 34 then	-- byte == '\"'
		local startOffset = dataTable.offset+1
		local endOffset = startOffset + 1
		while true do
			if endOffset > #dataTable.tag then
				TagError("Missing closing quote")
				dataTable.error = true
				return ""
			end
			byte = dataTable.tag:byte(endOffset) or 0
			if byte == 34 then -- byte == '\"'
				break
			elseif byte == 92 then	-- byte == '\\'
				endOffset = endOffset + 2
			else
				endOffset = endOffset + 1
			end
		end
		dataTable.offset = endOffset+1
		if functionString then
			return string.format("%s and %q", functionString, dataTable.tag:sub(startOffset, endOffset-1))
		else
			return string.format("%q", dataTable.tag:sub(startOffset, endOffset-1))
		end
		
	elseif (byte >= 65 and byte <= 90) or (byte >= 97 and byte <= 122) then	-- byte >= 'A' && byte <= 'Z' || byte >= 'a' && byte <= 'z'
		local startOffset = dataTable.offset
		local endOffset = startOffset
		while (byte >= 65 and byte <= 90) or (byte >= 97 and byte <= 122) do
			endOffset = endOffset + 1
			byte = dataTable.tag:byte(endOffset) or 0
		end
		dataTable.offset = endOffset
		local identifier = dataTable.tag:sub(startOffset, endOffset-1)
		local data = AloftTagData[identifier]
		if data then
            if data.events then
				if type(data.events) == "string" then
                    dataTable.events[data.events] = true
                end
            end
            if data.data then
                if type(data.data) == "string" then
                    dataTable.data[data.data] = true
                end
            end
		end
		if AloftTagDirectAccessData[identifier] then
			if functionString then
				return functionString.." and aloftData."..identifier
			else
				return "aloftData."..identifier
			end
		elseif AloftTagMethods[identifier] then
			dataTable.usesTagMethods = true
			local parameterList = self:ParseParameterList(dataTable)
			return string.format("tagMethods[\"%s\"](aloftData, %s%s)", identifier, functionString or "true", parameterList)
		else
			TagError("Invalid tag: "..identifier)
			dataTable.error = true
			return ""
		end
	end
end

function AloftTag:ParseChain(dataTable)
	local functionString = self:ParsePrimary(dataTable)
	while dataTable:GetByte() == 58 do	-- while tag[offset] == ':'
		dataTable:AdvanceByte()
		functionString = self:ParsePrimary(dataTable, functionString)
	end
	return functionString
end

function AloftTag:ParseExpression(dataTable)
	return self:ParseChain(dataTable)
end

function AloftTag:ParseRelation(dataTable)
	local result = self:ParseExpression(dataTable)

	local byte = dataTable:GetByte()
	local byte2 = dataTable.tag:byte(dataTable.offset+1)
	if byte == 61 then	-- '='
		dataTable:AdvanceByte()
		if byte2 == 61 then
			dataTable:AdvanceByte()
			return result.."=="..self:ParseExpression(dataTable)
		else
			TagError("Unexpected \"=\"")
			dataTable.error = true
			return ""
		end
	elseif byte == 126 then -- '~'
		dataTable:AdvanceByte()
		if byte2 == 61 then
			dataTable:AdvanceByte()
			return result.."~="..self:ParseExpression(dataTable)
		else
			TagError("Unexpected \"=\"")
			dataTable.error = true
			return ""
		end
	elseif byte == 60 then -- '<'
		dataTable:AdvanceByte()
		if byte2 == 61 then
			dataTable:AdvanceByte()
			return result.."<="..self:ParseExpression(dataTable)
		else
			return result.."<"..self:ParseExpression(dataTable)
		end
	elseif byte == 62 then -- '>'
		dataTable:AdvanceByte()
		if byte2 == 61 then
			dataTable:AdvanceByte()
			return result..">="..self:ParseExpression(dataTable)
		else
			return result..">"..self:ParseExpression(dataTable)
		end
	else
		return result
	end
end

function AloftTag:ParseRValue(dataTable)
	return self:ParseRelation(dataTable)
end


function AloftTag:CreateTag(tag)
	local dataTable = { data = { }, events = { }, tag = tag, offset = 1, GetByte = DataTableGetByte, AdvanceByte = DataTableAdvanceByte }

	if not tag or tag == "" then
		dataTable.method = EmptyStringMethod
		return dataTable
	end

	local functionStringList = { "",
								 "return function(aloftData) return " }

	while dataTable.offset <= #dataTable.tag and not dataTable.error do
		if dataTable.offset > 1 then
			table.insert(functionStringList, "..")
		end
		if dataTable.tag:byte(dataTable.offset) == 91 then	-- if tag[offset] == '['
			dataTable:AdvanceByte()
			table.insert(functionStringList, string.format([[(%s or "")]], self:ParseRValue(dataTable)))
			if dataTable:GetByte() == 93 then -- if tag[offset] == ']'
				dataTable:AdvanceByte()
			elseif not dataTable.error then
				TagError("Tag not closed - missing \"]\"")
				dataTable.error = true
			end
		else
			-- literal section
			local startOffset = dataTable.offset
			local endOffset = startOffset+1
			while endOffset <= #dataTable.tag and dataTable.tag:byte(endOffset) ~= 91 do
				endOffset = endOffset + 1
			end
			table.insert(functionStringList, string.format("%q", dataTable.tag:sub(startOffset, endOffset-1)))
			dataTable.offset = endOffset
		end
	end

	local functionString
	if dataTable.error then
		functionString = [[return function() return "Tag Error" end]]
	else
		if dataTable.usesTagMethods then
			functionStringList[1] = "local tagMethods = AloftTagMethods "
		end
		table.insert(functionStringList, " end")
		functionString = table.concat(functionStringList)
	end
	
--[[
	AloftDebug(string.format("Tag: %s", dataTable.tag))
	for event in pairs(dataTable.events) do
		AloftDebug(string.format(" - event: %s", event))
	end
	for data in pairs(dataTable.data) do
		AloftDebug(string.format(" - data: %s", data))
	end
	AloftDebug("Compiled method: "..functionString)
--]]--
	dataTable.method = loadstring(functionString)()

	return dataTable
end

-----------------------------------------------------------------------------
